from abc import ABC, abstractmethod

class Solver(ABC):

    @abstractmethod
    def solve(self):
        """
        Solves the equation with the chosen solver and returns the solution
        """
        pass



import numpy as np
import pandas as pd
import pdesolvers as pde

def main():

    # testing for heat equation

    # equation1 = (pde.HeatEquation(1, 100,30,10000, 0.01)
    #             .set_initial_temp(lambda x: np.sin(np.pi * x) + 5)
    #             .set_left_boundary_temp(lambda t: 20 * np.sin(np.pi * t) + 5)
    #             .set_right_boundary_temp(lambda t: t + 5))
    #
    #
    # solver1 = pde.Heat1DCNSolver(equation1)
    # solver2 = pde.Heat1DExplicitSolver(equation1)

    # testing for monte carlo pricing

    ticker = 'AAPL'

    # STOCK
    historical_data = pde.HistoricalStockData(ticker)
    historical_data.fetch_stock_data( "2024-03-21","2025-03-21")
    sigma, r = historical_data.estimate_metrics()
    current_price = historical_data.get_latest_stock_price()

    equation2 = pde.BlackScholesEquation(pde.OptionType.EUROPEAN_CALL, current_price, 100, r, sigma, 1, 100, 20000)

    solver1 = pde.BlackScholesCNSolver(equation2)
    solver2 = pde.BlackScholesExplicitSolver(equation2)
    sol1 = solver1.solve()
    sol1.plot()

    # COMPARISON
    #  look to see the corresponding option price for the expiration date and strike price
    pricing_1 = pde.BlackScholesFormula(pde.OptionType.EUROPEAN_CALL, current_price, 100, r, sigma, 1)
    pricing_2 = pde.MonteCarloPricing(pde.OptionType.EUROPEAN_CALL, current_price, 100, r, sigma, 1, 365, 1000)

    bs_price = pricing_1.get_black_scholes_merton_price()
    monte_carlo_price = pricing_2.get_monte_carlo_option_price()

    pde_price = sol1.get_result()[-1, 0]
    print(f"PDE Price: {pde_price}")
    print(f"Black-Scholes Price: {bs_price}")
    print(f"Monte-Carlo Price: {monte_carlo_price}")

if __name__ == "__main__":
    main()


import time
import logging

import numpy as np
import pdesolvers.solution as sol
import pdesolvers.pdes.heat_1d as heat
import pdesolvers.utils.utility as utility

from scipy.sparse.linalg import spsolve
from pdesolvers.solvers.solver import Solver

logging.basicConfig(
    level = logging.INFO,
    format = "{asctime} - {levelname} - {message}",
    style="{",
    datefmt="%Y-%m-%d %H:%M",
)

class Heat1DExplicitSolver(Solver):
    def __init__(self, equation: heat.HeatEquation):
        self.equation = equation

    def solve(self):
        """
        This method solves the heat (diffusion) equation using the explicit finite difference method

        :return: the solver instance with the computed temperature values
        """

        logging.info(f"Starting {self.__class__.__name__} with {self.equation.x_nodes+1} spatial nodes and {self.equation.t_nodes+1} time nodes.")

        start = time.perf_counter()

        x = self.equation.generate_grid(self.equation.length, self.equation.x_nodes)
        t = self.equation.generate_grid(self.equation.time, self.equation.t_nodes)

        dx = x[1] - x[0]
        dt_max = 0.5 * (dx**2) / self.equation.k

        if self.equation.t_nodes is None:
            dt = 0.8 * dt_max
            self.equation.t_nodes = int(self.equation.time/dt)
            dt = self.equation.time / self.equation.t_nodes
        else:
            dt = t[1] - t[0]

        if dt > dt_max:
            raise ValueError("User-defined t nodes is too small and exceeds the CFL condition. Possible action: Increase number of t nodes for stability!")

        u = np.zeros((self.equation.t_nodes, self.equation.x_nodes))

        u[0, :] = self.equation.get_initial_temp(x)
        u[:, 0] = self.equation.get_left_boundary(t)
        u[:, -1] = self.equation.get_right_boundary(t)

        for tau in range(0, self.equation.t_nodes-1):
            for i in range(1, self.equation.x_nodes - 1):
                u[tau+1,i] = u[tau, i] + (dt * self.equation.k * (u[tau, i-1] - 2 * u[tau, i] + u[tau, i+1]) / dx**2)

        end = time.perf_counter()
        duration = end - start

        logging.info(f"Solver completed in {duration} seconds.")

        return sol.Solution1D(u, x, t, dx, dt, duration)

class Heat1DCNSolver(Solver):
    def __init__(self, equation: heat.HeatEquation):
        self.equation = equation

    def solve(self):
        """
        This method solves the heat (diffusion) equation using the Crank Nicolson method

        :return: the solver instance with the computed temperature values
        """

        logging.info(f"Starting {self.__class__.__name__} with {self.equation.x_nodes+1} spatial nodes and {self.equation.t_nodes+1} time nodes.")

        start = time.perf_counter()

        x = self.equation.generate_grid(self.equation.length, self.equation.x_nodes)
        t = self.equation.generate_grid(self.equation.time, self.equation.t_nodes)

        dx = x[1] - x[0]
        dt = t[1] - t[0]

        alpha = self.equation.k * dt / (2 * dx**2)
        a = -alpha
        b = 1 + 2 * alpha
        c = -alpha

        u = np.zeros((self.equation.t_nodes, self.equation.x_nodes))

        u[0, :] = self.equation.get_initial_temp(x)
        u[:, 0] = self.equation.get_left_boundary(t)
        u[:, -1] = self.equation.get_right_boundary(t)

        lhs = utility.Heat1DHelper._build_tridiagonal_matrix(a, b, c, self.equation.x_nodes - 2)
        rhs = np.zeros(self.equation.x_nodes - 2)

        for tau in range(0, self.equation.t_nodes - 1):
            rhs[0] = alpha * (u[tau, 0] + u[tau+1, 0]) + (1 - 2 * alpha) * u[tau, 1] + alpha * u[tau, 2]

            for i in range(1, self.equation.x_nodes - 2):
                rhs[i] = alpha * u[tau, i] + (1 - 2 * alpha) * u[tau, i+1] + alpha * u[tau, i+2]

            rhs[-1] = alpha * (u[tau, -1] + u[tau+1, -1]) + (1 - 2 * alpha) * u[tau, -2] + alpha * u[tau, -3]

            u[tau+1, 1:-1] = spsolve(lhs, rhs)

        end = time.perf_counter()
        duration = end - start

        logging.info(f"Solver completed in {duration} seconds.")

        return sol.Solution1D(u, x, t, dx, dt, duration)


import numpy as np
import matplotlib.pyplot as plt
import time
import logging

from pdesolvers.enums.enums import OptionType

logging.basicConfig(
    level = logging.INFO,
    format = "{asctime} - {levelname} - {message}",
    style="{",
    datefmt="%Y-%m-%d %H:%M",
)

class MonteCarloPricing:

    def __init__(self, option_type: OptionType, S0, strike_price, mu, sigma, T, time_steps, sim, seed=None):
        """
        Initialize the Geometric Brownian Motion model with the given parameters.

        Parameters:
        - S0: Initial stock price
        - mu: Drift coefficient (expected return)
        - sigma: Volatility coefficient (standard deviation of returns)
        - T: Time period for the simulation (in years)
        - time_steps: Number of time steps in the simulation
        - sim: Number of simulations to run
        """

        self.__option_type = option_type
        self.__S0 = S0
        self.__strike_price = strike_price
        self.__mu = mu
        self.__sigma = sigma
        self.__T = T
        self.__time_steps = time_steps
        self.__sim = sim
        self.__seed = seed
        self.__S = None
        self.__payoff = None
        self.__run = False
        self.__duration = 0.0

    def get_monte_carlo_option_price(self):
        """
        Calculates the price of the option based on the monte carlo simulations
        :return: the option price
        """

        logging.info(f"Starting {self.__class__.__name__}")
        logging.info(f"Calculating the option price with {self.__sim} simulations")

        S = self.simulate_gbm()
        self.__run = True

        if self.__option_type == OptionType.EUROPEAN_CALL:
            self.__payoff = np.maximum(S[:, -1] - self.__strike_price, 0)
        elif self.__option_type == OptionType.EUROPEAN_PUT:
            self.__payoff = np.maximum(self.__strike_price - S[:, -1], 0)
        else:
            raise ValueError(f'Unsupported Option Type: {self.__option_type}')

        option_price = np.exp(-self.__mu * self.__T) * np.mean(self.__payoff)

        logging.info(f"Option price calculated successfully with {self.__class__.__name__}.")

        return option_price

    def simulate_gbm(self):
        """
        Simulate the Geometric Brownian Motion for the given parameters.

        This method calculates the stock prices at each time step for each simulation.
        """

        logging.info("Simulating the Geometric Brownian motion...")

        if self.__seed:
            np.random.seed(self.__seed)

        start = time.perf_counter()

        t = self.__generate_grid()
        dt = t[1] - t[0]

        B = np.zeros((self.__sim, self.__time_steps))
        S = np.zeros((self.__sim, self.__time_steps))

        # for all simulations at t = 0
        S[:,0] = self.__S0
        Z = np.random.normal(0, 1, (self.__sim, self.__time_steps))

        for i in range(self.__sim):
            for j in range (1, self.__time_steps):
                # updates brownian motion
                B[i,j] = B[i,j-1] + np.sqrt(dt) * Z[i,j-1]
                # calculates stock price based on the incremental difference
                S[i,j] = S[i, j-1] * np.exp((self.__mu - 0.5*self.__sigma**2)*dt + self.__sigma * (B[i, j] - B[i, j - 1]))

        end = time.perf_counter()
        self.__duration = end - start

        logging.info(f"Simulation completed in {self.__duration} seconds.")

        self.__S = S
        return self.__S

    def get_benchmark_errors(self, analytical_solution, num_simulations_list, export=False):
        """
        Calculates the absolute errors of each simulation count

        :param analytical_solution: the analytical solution of the black-scholes formula
        :param num_simulations_list: list of number of simulations
        :param export: boolean flag to indicate whether the results should be exported as csv
        :return: list of absolute errors corresponding to the number of simulations
        """

        prices = []
        errors = []

        original_sim = self.__sim

        for sim in num_simulations_list:
            self.__sim = sim
            self.__run = False

            print(f"Running simulation for Simulation Count: {sim}")

            sim_price = self.get_monte_carlo_option_price()
            prices.append(sim_price)
            error = abs(sim_price - analytical_solution)
            errors.append(error)

        self.__sim = original_sim
        self.__run = False

        if export:
            np.savetxt("benchmark_errors.csv", np.column_stack((num_simulations_list, errors)),
                       delimiter=",", header="Simulations,Error", comments="")

        return errors

    def __generate_grid(self):
        """
        Generate a time grid from 0 to T with `time_steps` intervals.

        Returns:
        - A numpy array representing the time grid.
        """

        return np.linspace(0, self.__T, self.__time_steps)

    def plot_price_paths(self, closing_prices=None, export=False):
        """
        Plot the simulated stock prices for all simulations.
        """

        if not self.__run:
            raise RuntimeError("Plots cannot be generated because the simulation has not been run yet.")

        plt.rcParams['font.family'] = 'monospace'
        plt.rcParams['font.size'] = 10

        t = self.__generate_grid()

        fig = plt.figure(figsize=(10,6))
        for i in range(np.min([100, self.__sim])):
            plt.plot(t, self.__S[i], color='grey', alpha=0.3)

        plt.title("Simulated Geometric Brownian Motion")
        plt.xlabel("Time (Years)")
        plt.ylabel("Stock Price")

        if closing_prices is not None:
            if len(closing_prices) != len(t):
                raise ValueError("Length of closing prices does not match the number of time steps in the simulation.")

            plt.plot(t , closing_prices, color='red')

        if export:
            plt.savefig("monte_carlo_prices.pdf", format="pdf", bbox_inches="tight")

        plt.show()

    def plot_distribution_of_final_prices(self, export=False):
        """
        Plots the distribution of final prices at expiry
        :param export: boolean flag to indicate whether the plot should be exported as pdf
        :return:
        """

        if not self.__run:
            raise RuntimeError("Plots cannot be generated because the simulation has not been run yet.")

        plt.rcParams['font.family'] = 'monospace'
        plt.rcParams['font.size'] = 10

        final_prices = self.__S[:, -1]

        plt.figure(figsize=(10, 6))
        plt.hist(final_prices, bins=50, edgecolor='darkblue', alpha=0.5, color='blue', density=True)
        plt.title('Distribution of Final Stock Prices at Maturity')
        plt.xlabel('Stock Price')
        plt.ylabel('Frequency Density')

        if export:
            plt.savefig("monte_carlo_prices.pdf", format="pdf", bbox_inches="tight", pad_inches=0.2)
            plt.show()

    def plot_distribution_of_payoff(self, export=False):
        """
        Plots the distribution of payoff at expiru
        :param export: boolean flag to indicate whether the plot should be exported as pdf
        :return: plot
        """
        if not self.__run:
            raise RuntimeError("Plots cannot be generated because the simulation has not been run yet.")

        plt.rcParams['font.family'] = 'monospace'
        plt.rcParams['font.size'] = 10

        plt.figure(figsize=(10, 6))
        plt.hist(self.__payoff, bins=50, edgecolor='darkblue', alpha=0.5, color='blue', density=True)
        plt.title('Distribution of Payoff')
        plt.xlabel('Payoff')
        plt.ylabel('Frequency Density')

        if export:
            plt.savefig("monte_carlo_payoff.pdf", format="pdf", bbox_inches="tight", pad_inches=0.2)

        plt.show()

    def plot_convergence_analysis(self, analytical_solution, num_simulations_list=None, export=False):
        """

        :param analytical_solution: the analytical solution of the black-scholes formula
        :param num_simulations_list: list of number of simulations
        :param export: boolean flag to indicate whether the plot should be exported as pdf
        :return: plot
        """
        if num_simulations_list is None:
            raise ValueError("Number of simulations need to be defined.")

        errors = self.get_benchmark_errors(analytical_solution, num_simulations_list, export=export)

        plt.rcParams['font.family'] = 'monospace'
        plt.rcParams['font.size'] = 10

        fig, ax = plt.subplots(figsize=(10,6))
        ax.plot(num_simulations_list, errors, 'bo-', linewidth=2)
        ax.set_xscale('log')
        ax.set_yscale('log')

        ax.set_xlabel('Number of Simulations')
        ax.set_ylabel('Absolute Error')
        ax.set_title(f'Monte Carlo Error Convergence (Benchmark: {analytical_solution:.4f})')

        if export:
            plt.savefig("monte_carlo_convergence.pdf", format="pdf",
                    bbox_inches="tight", pad_inches=0.2)

        plt.show()

    def get_execution_time(self):
        if not self.__run:
            raise RuntimeError("Execution time is not available because the simulation has not been run yet.")

        return self.__duration


import numpy as np
from scipy.stats import norm
from pdesolvers.enums.enums import OptionType

import logging

logging.basicConfig(
    level = logging.INFO,
    format = "{asctime} - {levelname} - {message}",
    style="{",
    datefmt="%Y-%m-%d %H:%M",
)

class BlackScholesFormula:
    def __init__(self, option_type: OptionType, S0, strike_price, r, sigma, expiry):
        self.__option_type = option_type
        self.__S0 = S0
        self.__strike_price = strike_price
        self.__r = r
        self.__sigma = sigma
        self.__expiry = expiry

    def get_black_scholes_merton_price(self):

        logging.info(f"Calculating the option price with analytical {self.__class__.__name__}")

        d1 = (np.log(self.__S0/self.__strike_price) + (self.__r + 0.5 * self.__sigma**2) * self.__expiry) / (self.__sigma * np.sqrt(self.__expiry))
        d2 = d1 - (self.__sigma * np.sqrt(self.__expiry))

        if self.__option_type == OptionType.EUROPEAN_CALL:
            price = self.__S0 * norm.cdf(d1) - self.__strike_price * np.exp(-self.__r * self.__expiry) * norm.cdf(d2)
        elif self.__option_type == OptionType.EUROPEAN_PUT:
            price =  self.__strike_price * np.exp(-self.__r * self.__expiry) * norm.cdf(-d2) - self.__S0 * norm.cdf(-d1)
        else:
            raise ValueError(f'Unsupported option type: {self.__option_type}')

        logging.info(f"Option price calculated successfully with {self.__class__.__name__}")

        return price


import numpy as np
import yfinance as yf

class HistoricalStockData:
    def __init__(self, ticker):
        self.__stock_data = None
        self.__ticker = ticker

    def fetch_stock_data(self, start_date, end_date):
        self.__stock_data = yf.download(self.__ticker, start=start_date, end=end_date, interval='1d')
        return self.__stock_data

    def estimate_metrics(self):

        if self.__stock_data is None:
            raise ValueError(f'No data available - data must be fetched first.')

        self.__stock_data["Log Returns"] = np.log(self.__stock_data["Close"] / self.__stock_data["Close"].shift(1))

        # takes into account the gaps between trading days
        self.__stock_data["Time Diff"] = self.__stock_data.index.to_series().diff().dt.days
        self.__stock_data["Z"] = self.__stock_data["Log Returns"] / self.__stock_data["Time Diff"]

        sigma = self.__stock_data["Log Returns"].std() * np.sqrt(252)
        mu = self.__stock_data["Log Returns"].mean() * 252

        return sigma, mu

    def get_latest_stock_price(self):
        if self.__stock_data is None:
            raise ValueError("No data available. Call fetch_data first.")

        return self.__stock_data["Close"].iloc[-1].item()

    def get_initial_stock_price(self):
        if self.__stock_data is None:
            raise ValueError("No data available. Call fetch_data first.")

        return self.__stock_data["Close"].iloc[0].item()

    def get_closing_prices(self):
        if self.__stock_data is None:
            raise ValueError("No data available. Call fetch_data first.")

        return self.__stock_data["Close"].values


import numpy as np
import pdesolvers.utils.utility as utility
import pdesolvers.enums.enums as enum

from matplotlib import pyplot as plt

import logging

logging.basicConfig(
    level = logging.INFO,
    format = "{asctime} - {levelname} - {message}",
    style="{",
    datefmt="%Y-%m-%d %H:%M",
)

class Solution:

    def __init__(self, result, x_grid, y_grid, dx, dy, duration):
        self.result = result
        self.x_grid = x_grid
        self.y_grid = y_grid
        self.dx = dx
        self.dy = dy
        self.duration = duration

    def plot(self, export=False):
        """
        Generates a 3D surface plot of the option values across a grid of asset prices and time

        :return: 3D surface plot
        """

        plt.rcParams['font.family'] = 'monospace'
        plt.rcParams['font.size'] = 10

        X, Y = np.meshgrid(self.x_grid, self.y_grid)

        # plotting the 3d surface
        fig = plt.figure(figsize=(10,6))
        ax = fig.add_subplot(111, projection='3d')
        surf = ax.plot_surface(X, Y, self.result, cmap='plasma', alpha=0.8)

        self._set_plot_labels(ax)

        if export:
            plt.savefig("solution_results.pdf", format="pdf", bbox_inches="tight", pad_inches=0.5)

        plt.show()

    def get_result(self):
        """
        Gets the grid of computed temperature values

        :return: grid result
        """
        return self.result

    def _set_plot_labels(self, ax):
        """
        Sets labels for the plot
        :param ax: axes
        """
        # font = {'family': 'serif'}
        ax.set_xlabel('X-axis')
        ax.set_ylabel('Y-axis')
        ax.set_zlabel('Value')
        ax.set_title('3D Surface Plot')

    def get_execution_time(self):
        """
        Gets the time taken for the solver to solve the equation
        :return: duration
        """
        return self.duration

    def __sub__(self, other):
        """
        Compares two solutions by interpolating the sparse grid to the dense grid and computing the difference
        :param other: the grid to be compared against
        :return: the error (difference) between the two solutions
        """

        logging.info("Interpolating solutions with RBFInterpolator")

        sparser_grid = other
        denser_grid = self

        if self.x_grid.shape[0] < other.x_grid.shape[0] and self.y_grid.shape[0] < other.y_grid.shape[0]:
            sparser_grid = self
            denser_grid = other

        interpolator_sparse = utility.RBFInterpolator(sparser_grid.result.T, sparser_grid.dx, sparser_grid.dy)

        diff = 0

        for idx_x in range(denser_grid.x_grid.shape[0]):
            for idx_y in range(denser_grid.y_grid.shape[0]):

                # Points (x, y) of the dense grid
                x = denser_grid.x_grid[idx_x]
                y = denser_grid.y_grid[idx_y]

                # Value at (x, y) for the dense grid
                val_dense_x_y = denser_grid.result[idx_y, idx_x]

                # Interpolate the sparse grid at (x, y)
                val_sparse_x_y = interpolator_sparse.interpolate(x, y)

                diff = np.max([diff, np.abs(val_dense_x_y - val_sparse_x_y)])

        logging.info("Interpolation completed.")

        return diff

class Solution1D(Solution):

    def __init__(self, result, x_grid, y_grid, dx, dy, duration):
        super().__init__(result, x_grid, y_grid, dx, dy, duration)

    def _set_plot_labels(self, ax):
        super()._set_plot_labels(ax)
        ax.set_xlabel('Space')
        ax.set_ylabel('Time')
        ax.set_zlabel('Temperature')
        ax.set_title('3D Surface Plot of 1D Heat Equation')


class SolutionBlackScholes(Solution):
    def __init__(self, result, x_grid, y_grid, dx, dy, duration, delta, gamma, theta, option_type):
        super().__init__(result, x_grid, y_grid, dx, dy, duration)
        self.option_type = option_type
        self.delta = delta
        self.gamma = gamma
        self.theta = theta

    def plot_greek(self, greek_type=enum.Greeks.DELTA, time_step=0, export=False):

        greek_types = {
            enum.Greeks.DELTA : {'data': self.delta, 'title': enum.Greeks.DELTA.value},
            enum.Greeks.GAMMA : {'data': self.gamma, 'title': enum.Greeks.GAMMA.value},
            enum.Greeks.THETA : {'data': self.theta, 'title': enum.Greeks.THETA.value}
        }

        # if greek_type.lower() not in greek_types:
        #     raise ValueError("Invalid greek type - please choose between delta/gamma/theta.")

        chosen_greek = greek_types[greek_type]
        greek_data = chosen_greek['data'][:, time_step]
        plt.figure(figsize=(8, 6))
        plt.plot(self.y_grid, greek_data, label=f"{chosen_greek['title']} at t={self.x_grid[time_step]:.4f}", color="grey", alpha=0.8)

        plt.title(f"{chosen_greek['title']} vs. Stock Price at t={self.x_grid[time_step]:.4f}")
        plt.xlabel("Stock Price (S)")
        plt.ylabel(chosen_greek['title'])
        plt.grid()
        plt.legend()

        if export:
            np.savetxt(f"option_{chosen_greek['title']}.csv", np.column_stack((self.y_grid, greek_data)), delimiter=',', header=f"StockPrice,{chosen_greek['title']}", comments='')

        plt.show()

    def _set_plot_labels(self, ax):
        super()._set_plot_labels(ax)
        ax.set_xlabel('Time')
        ax.set_ylabel('Asset Price')
        ax.set_zlabel(f'{self.option_type.value} Option Value')
        ax.set_title(f'{self.option_type.value} Option Value Surface Plot')


import pytest
import numpy as np

import pdesolvers.pdes.heat_1d as heat
import pdesolvers.solvers.heat_solvers as solver
import pdesolvers.utils.utility as utility

class TestHeatSolvers:

    def setup_method(self):
        self.equation = (heat.HeatEquation(1, 100,30,10000, 0.01)
                 .set_initial_temp(lambda x: 10 * np.sin(2 * np.pi * x) + 15)
                 .set_left_boundary_temp(lambda t: t + 15)
                 .set_right_boundary_temp(lambda t: 15))

    def test_check_terminal_and_boundary_conditions_at_time_zero(self):
        length = self.equation.length
        assert abs(self.equation.get_left_boundary(0) - self.equation.get_initial_temp(0)) < 1e-12, "Left boundary condition failed"
        assert abs(self.equation.get_right_boundary(0) - self.equation.get_initial_temp(length)) < 1e-12, "Right boundary condition failed"

    # explicit method tests

    def test_check_absolute_difference_between_two_results(self):
        result1 = solver.Heat1DExplicitSolver(self.equation).solve().get_result()
        result2 = solver.Heat1DCNSolver(self.equation).solve().get_result()

        diff = result1 - result2

        assert np.max(np.abs(diff)) < 1e-2

    def test_convergence_between_single_interpolated_point(self):
        result1 = solver.Heat1DExplicitSolver(self.equation).solve()
        result2 = solver.Heat1DCNSolver(self.equation).solve()
        u1 = result1.get_result()
        u2 = result2.get_result()

        data1 = utility.RBFInterpolator(u1, 0.1, 0.03).interpolate(0.2,0.9)
        data2 = utility.RBFInterpolator(u2, 0.1, 0.03).interpolate(0.2,0.9)

        diff = np.abs(data1 - data2)

        assert diff < 1e-4

    def test_convergence_between_two_interpolated_grids(self):
        result1 = solver.Heat1DExplicitSolver(self.equation).solve()
        result2 = solver.Heat1DCNSolver(self.equation).solve()

        diff = np.abs(result1 - result2)

        assert diff < 1e-1




import numpy as np
from pdesolvers.enums.enums import OptionType

class BlackScholesEquation:

    def __init__(self, option_type: OptionType, S_max, K, r, sigma, expiry, s_nodes=1, t_nodes=None):
        """
        Initialises the solver with the necessary parameters

        :param option_type: the type of option
        :param S_max: maximum asset price in the grid
        :param K: strike price
        :param r: risk-free interest rate
        :param sigma: volatility of the asset
        :param expiry: time to maturity/expiry of the option
        :param s_nodes: number of asset price nodes
        :param t_nodes: number of time nodes
        """

        if not isinstance(option_type, OptionType):
            raise TypeError(f"Option type must be of type OptionType enum" )
        self.__option_type = option_type
        self.__S_max = S_max
        self.__expiry = expiry
        self.__sigma = sigma
        self.__r = r
        self.__K = K
        self.__s_nodes = s_nodes
        self.__t_nodes = t_nodes
        self.__V = None

    @staticmethod
    def generate_grid(value, nodes):
        return np.linspace(0, value, nodes + 1)

    @property
    def s_nodes(self):
        return self.__s_nodes

    @property
    def t_nodes(self):
        return self.__t_nodes

    @property
    def option_type(self):
        return self.__option_type

    @property
    def S_max(self):
        return self.__S_max

    @property
    def sigma(self):
        return self.__sigma

    @property
    def expiry(self):
        return self.__expiry

    @property
    def rate(self):
        return self.__r

    @property
    def strike_price(self):
        return self.__K

    @t_nodes.setter
    def t_nodes(self, nodes):
        self.__t_nodes = nodes

    @option_type.setter
    def option_type(self, type):
        self.__option_type = type

    @S_max.setter
    def S_max(self, asset_price):
        self.__S_max = asset_price

    @expiry.setter
    def expiry(self, expiry):
        self.__expiry = expiry

    @sigma.setter
    def sigma(self, sigma):
        self.__sigma = sigma

    @rate.setter
    def rate(self, rate):
        self.__r = rate

    @strike_price.setter
    def strike_price(self, price):
        self.__K = price


import pytest
import numpy as np
from unittest.mock import patch

from pdesolvers.optionspricing.monte_carlo import MonteCarloPricing
from pdesolvers.enums.enums import OptionType, Greeks

@pytest.fixture
def mc_pricing_params():
    return {
        'S0': 100.0,
        'strike_price': 100.0,
        'mu': 0.05,
        'sigma': 0.2,
        'T': 1.0,
        'time_steps': 252,
        'sim': 5
    }

class TestMonteCarlo:

    @patch('pdesolvers.optionspricing.monte_carlo.MonteCarloPricing.simulate_gbm')
    def test_get_european_call_option_price(self, mock_simulate_gbm, mc_pricing_params):
        mock_price_array = np.zeros((5, mc_pricing_params['time_steps']))
        mock_price_array[:, -1] = np.array([110, 120, 90, 105, 115])

        mock_simulate_gbm.return_value = mock_price_array

        test_mc = MonteCarloPricing(
            OptionType.EUROPEAN_CALL, **mc_pricing_params
        )

        mock_payoffs = np.maximum(np.array([110, 120, 90, 105, 115]) - mc_pricing_params['strike_price'], 0)
        expected_price = np.exp(-mc_pricing_params['mu'] * mc_pricing_params['T']) * np.mean(mock_payoffs)

        actual_price = test_mc.get_monte_carlo_option_price()

        assert actual_price == expected_price

    @patch('pdesolvers.optionspricing.monte_carlo.MonteCarloPricing.simulate_gbm')
    def test_get_european_put_option_price(self, mock_simulate_gbm, mc_pricing_params):
        mock_price_array = np.zeros((5, mc_pricing_params['time_steps']))
        mock_price_array[:, -1] = np.array([110, 120, 90, 105, 115])

        mock_simulate_gbm.return_value = mock_price_array

        test_mc = MonteCarloPricing(
            OptionType.EUROPEAN_PUT, **mc_pricing_params
        )

        mock_payoffs = np.maximum(mc_pricing_params['strike_price'] - np.array([110, 120, 90, 105, 115]), 0)
        expected_price = np.exp(-mc_pricing_params['mu'] * mc_pricing_params['T']) * np.mean(mock_payoffs)

        actual_price = test_mc.get_monte_carlo_option_price()

        assert actual_price == expected_price

    def test_get_invalid_option_price(self, mc_pricing_params):

        test_mc = MonteCarloPricing(
            Greeks.THETA, **mc_pricing_params
        )

        with pytest.raises(ValueError, match="Unsupported Option Type: Greeks.THETA"):
            test_mc.get_monte_carlo_option_price()

    def test_check_simulate_gbm_results_shape(self, mc_pricing_params):

        test_mc = MonteCarloPricing(
            OptionType.EUROPEAN_PUT, **mc_pricing_params
        )

        results = test_mc.simulate_gbm()

        assert results.shape == (mc_pricing_params['sim'], mc_pricing_params['time_steps'])

    def test_check_simulate_gbm_initial_values(self, mc_pricing_params):
        test_mc = MonteCarloPricing(
            OptionType.EUROPEAN_CALL, **mc_pricing_params
        )

        results = test_mc.simulate_gbm()

        assert (results[:,0] == mc_pricing_params['S0']).all()

    def test_check_get_execution_time_raises_exception_when_no_simulation_is_run(self, mc_pricing_params):

        test_mc = MonteCarloPricing(
            OptionType.EUROPEAN_CALL, **mc_pricing_params
        )

        with pytest.raises(RuntimeError, match="Execution time is not available because the simulation has not been run yet."):
            test_mc.get_execution_time()



import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from scipy.sparse import csc_matrix

import pdesolvers.enums.enums as enum

class Heat1DHelper:

    @staticmethod
    def _build_tridiagonal_matrix(a, b, c, nodes):
        """
        Initialises the tridiagonal matrix on the LHS of the equation

        :param a: the coefficient of U @ (t = tau + 1 & x = i-1)
        :param b: the coefficient of U @ (t = tau + 1 & x = i)
        :param c: the coefficient of U @ (t = tau + 1 & x = i+1)
        :param nodes: number of spatial nodes ( used to initialise the size of the tridiagonal matrix)
        :return: the tridiagonal matrix consisting of coefficients
        """

        matrix = np.zeros((nodes, nodes))
        np.fill_diagonal(matrix, b)
        np.fill_diagonal(matrix[1:], a)
        np.fill_diagonal(matrix[:, 1:], c)

        matrix = csc_matrix(matrix)

        return matrix

class BlackScholesHelper:

    @staticmethod
    def _calculate_greeks_at_boundary(equation, delta, gamma, theta, tau, V, S, ds):
        delta[0, tau] = (V[1, tau+1] - V[0, tau+1]) / ds
        delta[equation.s_nodes, tau] = (V[equation.s_nodes, tau+1] - V[equation.s_nodes-1, tau+1]) / ds

        gamma[0, tau] = (V[2, tau+1] - 2*V[1, tau+1] + V[0, tau+1]) / (ds**2)
        gamma[equation.s_nodes, tau] = (V[equation.s_nodes, tau+1] - 2*V[equation.s_nodes-1, tau+1] + V[equation.s_nodes-2, tau+1]) / (ds**2)

        theta[0, tau] = -0.5 * (equation.sigma**2) * (S[0]**2) * gamma[0, tau] - equation.rate * S[0] * delta[0, tau] + equation.rate * V[0, tau+1]
        theta[equation.s_nodes, tau] = -0.5 * (equation.sigma**2) * (S[-1]**2) * gamma[equation.s_nodes, tau] - equation.rate * S[-1] * delta[equation.s_nodes, tau] + equation.rate * V[equation.s_nodes, tau+1]

        return delta, gamma, theta

    @staticmethod
    def _set_boundary_conditions(equation, T, tau):
        """
        Sets the boundary conditions for the Black-Scholes Equation based on option type

        :param T: grid of time steps
        :param tau: index of current time step
        :return: a tuple representing the boundary values for the given time step
        """

        lower_boundary = None
        upper_boundary = None
        if equation.option_type == enum.OptionType.EUROPEAN_CALL:
            lower_boundary = 0
            upper_boundary = equation.S_max - equation.strike_price * np.exp(-equation.rate * (equation.expiry - T[tau]))
        elif equation.option_type == enum.OptionType.EUROPEAN_PUT:
            lower_boundary = equation.strike_price * np.exp(-equation.rate * (equation.expiry - T[tau]))
            upper_boundary = 0

        return lower_boundary, upper_boundary

class RBFInterpolator:

    def __init__(self, z, hx, hy):
        """
        Initializes the RBF Interpolator.

        :param z: 2D array of values at the grid points.
        :param x: x-coordinate of the point to interpolate.
        :param y: y-coordinate of the point to interpolate.
        :param hx: Grid spacing in the x-direction.
        :param hy: Grid spacing in the y-direction.
        """

        self.__z = z
        self.__hx = hx
        self.__hy = hy
        self.__nx, self.__ny = z.shape

    def __get_coordinates(self, x, y):
        """
        Determines the x and y coordinates of the bottom-left corner of the grid cell

        :return: A tuple containing the coordinates and its corresponding indices
        """

        # gets the grid steps to x
        i_minus_star = int(np.floor(x / self.__hx))
        i_minus = min(max(0, i_minus_star), self.__nx - 2)

        # gets the grid steps to y
        j_minus_star = int(np.floor(y / self.__hy))
        j_minus = min(max(0, j_minus_star), self.__ny - 2)

        # computes the coordinates at the computed indices
        x_minus = i_minus * self.__hx
        y_minus = j_minus * self.__hy

        return x_minus, y_minus, i_minus, j_minus

    def __euclidean_distances(self, x_minus, y_minus, x, y):
        """
        Calculates Euclidean distances between (x,y) and the surrounding grid points in the unit cell

        :param x_minus: x-coordinate of the bottom-left corner of the grid
        :param y_minus: y-coordinate of the bottom-left corner of the grid
        :return: returns tuple with the Euclidean distances to the surrounding grid points:
                [bottom left, top left, bottom right, top right]
        """

        bottom_left = np.sqrt((x_minus - x) ** 2 + (y_minus - y) ** 2)
        top_left = np.sqrt((x_minus - x) ** 2 + (y_minus + self.__hy - y) ** 2)
        bottom_right = np.sqrt((x_minus + self.__hx - x) ** 2 + (y_minus - y) ** 2)
        top_right = np.sqrt((x_minus + self.__hx - x) ** 2 + (y_minus + self.__hy - y) ** 2)

        return bottom_left, top_left, bottom_right, top_right

    @staticmethod
    def __rbf(d, gamma):
        """
        Computes the Radial Basis Function (RBF) for a given distance and gamma

        :param d: the Euclidean distance to a grid point
        :param gamma: gamma parameter
        :return: the RBF value for the distance d
        """
        return np.exp(-gamma * d ** 2)

    def interpolate(self, x, y):
        """
        Performs the Radial Basis function (RBF) interpolation for the point (x,y)

        :return: the interpolated value at (x,y)
        """

        x_minus, y_minus, i_minus, j_minus = self.__get_coordinates(x, y)

        distances = self.__euclidean_distances(x_minus, y_minus, x, y)

        h_diag_squared = self.__hx ** 2 + self.__hy ** 2
        gamma = -np.log(0.005) / h_diag_squared

        rbf_weights = [self.__rbf(d, gamma) for d in distances]

        sum_rbf = np.sum(rbf_weights)
        interpolated = rbf_weights[0] * self.__z[i_minus, j_minus]
        interpolated += rbf_weights[1] * self.__z[i_minus, j_minus + 1]
        interpolated += rbf_weights[2] * self.__z[i_minus + 1, j_minus]
        interpolated += rbf_weights[3] * self.__z[i_minus + 1, j_minus + 1]
        interpolated /= sum_rbf

        return interpolated

class GPUResults:

    def __init__(self, file_path, s_max, expiry):
        self.__file_path = file_path
        self.__s_max = s_max
        self.__expiry = expiry
        self.__grid_data = None

    def get_results(self):

        # Load data
        df = pd.read_csv(self.__file_path, header=None)
        print(f"Data shape: {df.shape}")

        self.__grid_data = df.values.T

        return self.__grid_data

    def plot_option_surface(self):
        if self.__grid_data is None:
            self.get_results()

        price_grid = np.linspace(0, self.__s_max, self.__grid_data.shape[0])
        time_grid = np.linspace(0, self.__expiry, self.__grid_data.shape[1])
        X, Y = np.meshgrid(time_grid, price_grid)

        fig = plt.figure(figsize=(10, 6))
        ax = fig.add_subplot(111, projection='3d')
        surf = ax.plot_surface(X, Y, self.__grid_data, cmap='viridis')

        ax.set_xlabel('Time')
        ax.set_ylabel('Asset Price')
        ax.set_zlabel('Option Value')
        ax.set_title('Option Value Surface Plot')
        fig.colorbar(surf, shrink=0.5, aspect=5)

        plt.show()


import numpy as np
import pdesolvers.solution as sol
import pdesolvers.pdes.black_scholes as bse
import pdesolvers.enums.enums as enum
import pdesolvers.utils.utility as utility
import time
import logging

from scipy import sparse
from scipy.sparse.linalg import spsolve
from pdesolvers.solvers.solver import Solver

logging.basicConfig(
    level = logging.INFO,
    format = "{asctime} - {levelname} - {message}",
    style="{",
    datefmt="%Y-%m-%d %H:%M",
)

class BlackScholesExplicitSolver(Solver):

    def __init__(self, equation: bse.BlackScholesEquation):
        self.equation = equation


    def solve(self):
        """
        This method solves the Black-Scholes equation using the explicit finite difference method

        :return: the solver instance with the computed option values
        """


        logging.info(f"Starting {self.__class__.__name__} with {self.equation.s_nodes+1} spatial nodes and {self.equation.t_nodes+1} time nodes.")
        start = time.perf_counter()

        S = self.equation.generate_grid(self.equation.S_max, self.equation.s_nodes)
        T = self.equation.generate_grid(self.equation.expiry, self.equation.t_nodes)

        dt_max = 1/((self.equation.s_nodes**2) * (self.equation.sigma**2)) # cfl condition to ensure stability

        if self.equation.t_nodes is None:
            dt = 0.9 * dt_max
            self.equation.t_nodes = int(self.equation.expiry/dt)
            dt = self.equation.expiry / self.equation.t_nodes
        else:
            dt = T[1] - T[0]

            if dt > dt_max:
                logging.error(f'The supplied grid dimension do not satisfy the CFL condition.')
                raise ValueError("User-defined t nodes is too small and exceeds the CFL condition. Possible action: Increase number of t nodes for stability!")

        ds = S[1] - S[0]

        V = np.zeros((self.equation.s_nodes + 1, self.equation.t_nodes + 1))

        # setting terminal condition
        if self.equation.option_type == enum.OptionType.EUROPEAN_CALL:
            V[:,-1] = np.maximum((S - self.equation.strike_price), 0)
        elif self.equation.option_type == enum.OptionType.EUROPEAN_PUT:
            V[:,-1] = np.maximum((self.equation.strike_price - S), 0)
        else:
            raise ValueError("Invalid option type - please choose between call/put")

        delta = np.zeros((self.equation.s_nodes + 1, self.equation.t_nodes + 1))
        gamma = np.zeros((self.equation.s_nodes + 1, self.equation.t_nodes + 1))
        theta = np.zeros((self.equation.s_nodes + 1, self.equation.t_nodes + 1))

        for tau in reversed(range(self.equation.t_nodes)):
            for i in range(1, self.equation.s_nodes):
                delta[i, tau] = (V[i+1, tau+1] - V[i-1, tau+1]) / (2 * ds)
                gamma[i, tau] = (V[i+1, tau+1] - 2 * V[i,tau+1] + V[i-1, tau+1]) / (ds ** 2)
                theta[i, tau] = -0.5 * (self.equation.sigma ** 2) * (S[i] ** 2) * gamma[i, tau] - self.equation.rate * S[i] * delta[i, tau] + self.equation.rate * V[i, tau+1]
                V[i, tau] = V[i, tau + 1] - (theta[i, tau] * dt)

            # setting boundary conditions
            lower, upper = utility.BlackScholesHelper._set_boundary_conditions(self.equation, T, tau)
            V[0, tau] = lower
            V[self.equation.s_nodes, tau] = upper

            delta, gamma, theta = utility.BlackScholesHelper._calculate_greeks_at_boundary(self.equation, delta, gamma, theta, tau, V, S, ds)

        end = time.perf_counter()
        duration = end - start

        logging.info(f"Solver completed in {duration} seconds.")

        return sol.SolutionBlackScholes(V, T, S, dt, ds, duration, delta, gamma, theta, self.equation.option_type)


class BlackScholesCNSolver(Solver):

    def __init__(self, equation: bse.BlackScholesEquation):
        self.equation = equation

    def solve(self):
        """
        This method solves the Black-Scholes equation using the Crank-Nicolson method

        :return: the solver instance with the computed option values
        """

        logging.info(f"Starting {self.__class__.__name__} with {self.equation.s_nodes+1} spatial nodes and {self.equation.t_nodes+1} time nodes.")

        start = time.perf_counter()
        S = self.equation.generate_grid(self.equation.S_max, self.equation.s_nodes)
        T = self.equation.generate_grid(self.equation.expiry, self.equation.t_nodes)

        ds = S[1] - S[0]
        dt = T[1] - T[0]

        a = 0.25 * dt * ((self.equation.sigma**2) * (S**2) / (ds**2) - self.equation.rate * S / ds)
        b = -dt * 0.5 * (self.equation.sigma**2 * (S**2) / (ds**2) + self.equation.rate)
        c = 0.25 * dt * (self.equation.sigma**2 * (S**2) / (ds**2) + self.equation.rate * S / ds)

        lhs = sparse.diags([-a[2:], 1-b[1:], -c[1:-1]], [-1, 0, 1], shape = (self.equation.s_nodes - 1, self.equation.s_nodes - 1), format='csr')
        rhs = sparse.diags([a[2:], 1+b[1:], c[1:-1]], [-1, 0, 1], shape = (self.equation.s_nodes - 1, self.equation.s_nodes - 1) , format='csr')

        V = np.zeros((self.equation.s_nodes+1, self.equation.t_nodes+1))

        delta = np.zeros((self.equation.s_nodes + 1, self.equation.t_nodes + 1))
        gamma = np.zeros((self.equation.s_nodes + 1, self.equation.t_nodes + 1))
        theta = np.zeros((self.equation.s_nodes + 1, self.equation.t_nodes + 1))


        # setting terminal condition (for all values of S at time T)
        if self.equation.option_type == enum.OptionType.EUROPEAN_CALL:
            V[:,-1] = np.maximum((S - self.equation.strike_price), 0)

            # setting boundary conditions (for all values of t at asset prices S=0 and S=Smax)
            V[0, :] = 0
            V[-1, :] = S[-1] - self.equation.strike_price * np.exp(-self.equation.rate * (self.equation.expiry - T))

        elif self.equation.option_type == enum.OptionType.EUROPEAN_PUT:
            V[:,-1] = np.maximum((self.equation.strike_price - S), 0)
            V[0, :] = self.equation.strike_price * np.exp(-self.equation.rate * (self.equation.expiry - T))
            V[-1, :] = 0

        for tau in reversed(range(self.equation.t_nodes)):
            # Construct the RHS vector for this time step
            rhs_vector = rhs @ V[1:-1, tau + 1]

            # Apply boundary conditions to the RHS vector
            rhs_vector[0] += a[1] * (V[0, tau + 1] + V[0, tau])
            rhs_vector[-1] += c[self.equation.s_nodes-1] *(V[-1, tau+1] + V[-1, tau])

            # Solve the linear system for interior points
            V[1:-1, tau] = spsolve(lhs, rhs_vector)

            # Calculate Greeks for interior points
            delta[1:-1, tau] = (V[2:, tau] - V[:-2, tau]) / (2 * ds)
            gamma[1:-1, tau] = (V[2:, tau] - 2 * V[1:-1, tau] + V[:-2, tau]) / (ds**2)
            theta[1:-1, tau] = -0.5 * (self.equation.sigma**2) * (S[1:-1]**2) * gamma[1:-1, tau] - self.equation.rate * S[1:-1] * delta[1:-1, tau] + self.equation.rate * V[1:-1, tau]

            delta, gamma, theta = utility.BlackScholesHelper._calculate_greeks_at_boundary(self.equation, delta, gamma, theta, tau, V, S, ds)

        end = time.perf_counter()
        duration = end - start

        logging.info(f"Solver completed in {duration} seconds.")

        return sol.SolutionBlackScholes(V, T, S, dt, ds, duration, delta, gamma, theta, self.equation.option_type)




import numpy as np

class HeatEquation:

    def __init__(self, length, x_nodes, time, t_nodes, k):
        self.__length = length
        self.__x_nodes = x_nodes
        self.__time = time
        self.__t_nodes = t_nodes
        self.__k = k
        self.__initial_temp = None
        self.__left_boundary_temp = None
        self.__right_boundary_temp = None

    def set_initial_temp(self, u0):
        self.__validate_callable(u0)
        if self.__length is None:
            raise RuntimeError("Rod length has not been initialised.")
        self.__initial_temp = u0
        self.__check_conditions()
        return self

    def set_left_boundary_temp(self, left):
        self.__validate_callable(left)
        self.__left_boundary_temp = left
        self.__check_conditions()
        return self

    def set_right_boundary_temp(self, right):
        self.__validate_callable(right)
        self.__right_boundary_temp = right
        self.__check_conditions()
        return self

    def __check_conditions(self):
        if self.__initial_temp is None:
            raise ValueError("Initial Temperature has not been initialised")

        if self.__left_boundary_temp is not None:
            err = np.abs(self.__left_boundary_temp(0) - self.__initial_temp(0))
            assert err < 1e-12, f"Left boundary condition at t=0 does not match the initial condition."

        if self.__right_boundary_temp is not None:
            err = np.abs(self.__right_boundary_temp(0) - self.__initial_temp(self.__length))
            assert err < 1e-12, f"Right boundary condition at t=0 does not match the initial condition."

    @staticmethod
    def __validate_callable(func):
        if not callable(func):
            raise ValueError("Temperature conditions must be a callable function")

    @staticmethod
    def generate_grid(value, nodes):
        return np.linspace(0, value, nodes)

    @property
    def length(self):
        return self.__length

    @property
    def time(self):
        return self.__time

    @property
    def x_nodes(self):
        return self.__x_nodes

    @property
    def t_nodes(self):
        return self.__t_nodes

    @property
    def k(self):
        return self.__k

    def get_initial_temp(self, x):
        return self.__initial_temp(x)

    def get_left_boundary(self, t):
        return self.__left_boundary_temp(t)

    def get_right_boundary(self, t):
        return self.__right_boundary_temp(t)

    @t_nodes.setter
    def t_nodes(self, nodes):
        self.__t_nodes = nodes



from setuptools import setup, find_packages
import io
import os

here = os.path.abspath(os.path.dirname(__file__))

NAME = 'pdesolvers'

# Import version from file
version_file = open(os.path.join(here, 'VERSION'))
VERSION = version_file.read().strip()

DESCRIPTION = 'A package for solving partial differential equations'

# Import the README and use it as the long-description.
try:
    with io.open(os.path.join(here, 'README.md'), encoding='utf-8') as f:
        LONG_DESCRIPTION = '\n' + f.read()
except FileNotFoundError:
    long_description = DESCRIPTION

# Read requirements from requirements.txt
with open('requirements.txt') as f:
    requirements = [line.strip() for line in f if line.strip() and not line.startswith('#')]

setup(
    name=NAME,
    version=VERSION,
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    long_description_content_type='text/markdown',\
    author='Chelsea De Marseilla, Debdal Chowdhury',
    license='Apache License 2.0',
    packages=find_packages(),
    install_requires=requirements,
    extras_require={
        'package-tools': [
            'setuptools',
            'wheel',
            'twine'
        ]
    },
    url='https://github.com/GPUEngineering/PDESolvers',
)


import pytest
import numpy as np
import pdesolvers.pdes.black_scholes as bse
import pdesolvers.solvers.black_scholes_solvers as solver
import pdesolvers.utils.utility as utility

from pdesolvers.optionspricing.black_scholes_formula import BlackScholesFormula
from pdesolvers.enums.enums import OptionType, Greeks

@pytest.fixture
def bsf_pricing_params():
    return {
        'S0': 300.0,
        'strike_price': 100.0,
        'r': 0.05,
        'sigma': 0.2,
        'expiry': 1.0
    }


class TestBlackScholesEquation:

    def test_check_invalid_option_type_input(self):
        with pytest.raises(TypeError, match="Option type must be of type OptionType enum"):
            self.equation = bse.BlackScholesEquation('woo', 300, 100, 0.05, 0.2, 1, 100, 2000)

class TestBlackScholesSolvers:

    def setup_method(self):
        self.equation = bse.BlackScholesEquation(OptionType.EUROPEAN_CALL, 300, 100, 0.05, 0.2, 1, 100, 2000)
        self.test_asset_grid = self.equation.generate_grid(self.equation.S_max, self.equation.s_nodes)
        self.test_time_grid = self.equation.generate_grid(self.equation.expiry, self.equation.t_nodes)

    # explicit method tests

    def test_check_lower_boundary_for_call_explicit(self):
        result = solver.BlackScholesExplicitSolver(self.equation).solve().get_result()
        assert np.all(result[0,:]) == 0

    def test_check_upper_boundary_for_call_explicit(self):
        result = solver.BlackScholesExplicitSolver(self.equation).solve().get_result()

        expected_value = self.test_asset_grid[-1] - self.equation.strike_price * np.exp(-self.equation.rate * (self.equation.expiry - self.test_time_grid))
        assert np.isclose(result[-1, :], expected_value, atol=1e-6).all()

    def test_check_terminal_condition_for_call_explicit(self):
        result = solver.BlackScholesExplicitSolver(self.equation).solve().get_result()

        test_strike_price = self.equation.strike_price
        expected_payoff = np.maximum(self.test_asset_grid - test_strike_price, 0)

        assert np.array_equal(result[:, -1], expected_payoff)

    def test_check_lower_boundary_for_put_explicit(self):
        self.equation.option_type = OptionType.EUROPEAN_PUT
        result = solver.BlackScholesExplicitSolver(self.equation).solve().get_result()

        expected_value = self.equation.strike_price * np.exp(-self.equation.rate * (self.equation.expiry - self.test_time_grid))
        assert np.isclose(result[0, :], expected_value, atol=1e-6).all()

    def test_check_upper_boundary_for_put_explicit(self):
        self.equation.option_type = OptionType.EUROPEAN_PUT
        result = solver.BlackScholesExplicitSolver(self.equation).solve().get_result()

        assert np.all(result[-1,:]) == 0

    def test_check_terminal_condition_for_put_explicit(self):
        self.equation.option_type = OptionType.EUROPEAN_PUT
        result = solver.BlackScholesExplicitSolver(self.equation).solve().get_result()

        test_strike_price = self.equation.strike_price
        expected_payoff = np.maximum(test_strike_price - self.test_asset_grid, 0)

        assert np.array_equal(result[:,-1], expected_payoff)

    def test_check_valid_option_type(self):
        self.equation.option_type = "INVALID"

        with pytest.raises(ValueError, match="Invalid option type - please choose between call/put"):
            solver.BlackScholesExplicitSolver(self.equation).solve().get_result()

    # crank-nicolson method tests

    def test_check_lower_boundary_for_call_cn(self):
        result = solver.BlackScholesCNSolver(self.equation).solve().get_result()
        assert np.all(result[0,:]) == 0

    def test_check_upper_boundary_for_call_cn(self):
        result = solver.BlackScholesCNSolver(self.equation).solve().get_result()

        expected_value = self.test_asset_grid[-1] - self.equation.strike_price * np.exp(-self.equation.rate * (self.equation.expiry - self.test_time_grid))
        assert np.isclose(result[-1, :], expected_value, atol=1e-6).all()

    def test_check_terminal_condition_for_call_cn(self):
        result = solver.BlackScholesCNSolver(self.equation).solve().get_result()

        test_strike_price = self.equation.strike_price
        expected_payoff = np.maximum(self.test_asset_grid - test_strike_price, 0)

        assert np.array_equal(result[:, -1], expected_payoff)

    def test_check_lower_boundary_for_put_cn(self):
        self.equation.option_type = OptionType.EUROPEAN_PUT
        result = solver.BlackScholesCNSolver(self.equation).solve().get_result()

        expected_value = self.equation.strike_price * np.exp(-self.equation.rate * (self.equation.expiry - self.test_time_grid))
        assert np.isclose(result[0, :], expected_value, atol=1e-6).all()

    def test_check_upper_boundary_for_cn(self):
        self.equation.option_type = OptionType.EUROPEAN_PUT
        result = solver.BlackScholesCNSolver(self.equation).solve().get_result()

        assert np.all(result[-1,:]) == 0

    def test_check_terminal_condition_for_put_cn(self):
        self.equation.option_type = OptionType.EUROPEAN_PUT
        result = solver.BlackScholesCNSolver(self.equation).solve().get_result()

        test_strike_price = self.equation.strike_price
        expected_payoff = np.maximum(test_strike_price - self.test_asset_grid, 0)

        assert np.array_equal(result[:,-1], expected_payoff)

    def test_check_absolute_difference_between_two_results(self):
        result1 = solver.BlackScholesExplicitSolver(self.equation).solve()
        result2 = solver.BlackScholesCNSolver(self.equation).solve()
        u1 = result1.get_result()
        u2 = result2.get_result()
        diff = u1 - u2

        assert np.max(np.abs(diff)) < 1e-2

    def test_convergence_between_single_interpolated_point(self):
        result1 = solver.BlackScholesExplicitSolver(self.equation).solve()
        result2 = solver.BlackScholesCNSolver(self.equation).solve()
        u1 = result1.get_result()
        u2 = result2.get_result()

        data1 = utility.RBFInterpolator(u1, 0.1, 0.03).interpolate(0.2,0.9)
        data2 = utility.RBFInterpolator(u2, 0.1, 0.03).interpolate(0.2,0.9)

        diff = np.abs(data1 - data2)

        assert diff < 1e-4

    def test_convergence_between_two_interpolated_grids(self):
        result1 = solver.BlackScholesExplicitSolver(self.equation).solve()
        result2 = solver.BlackScholesCNSolver(self.equation).solve()

        diff = result1 - result2

        assert diff < 1e-1

    def test_convergence_between_same_solver_with_different_grid_sizes(self):
        temp_equation = bse.BlackScholesEquation(OptionType.EUROPEAN_CALL, 300, 100, 0.05, 0.2, 1, 100, 1000)

        result1 = solver.BlackScholesExplicitSolver(self.equation).solve()
        result2 = solver.BlackScholesExplicitSolver(temp_equation).solve()

        diff = result1 - result2
        print(diff)

        assert diff < 1e-1

    def test_compare_results_between_solver_and_analytical_formula(self, bsf_pricing_params):
        result1 = solver.BlackScholesCNSolver(self.equation).solve().get_result()
        test_bsf = BlackScholesFormula(
            OptionType.EUROPEAN_CALL, **bsf_pricing_params
        )

        asset_grid = self.equation.generate_grid(self.equation.S_max, self.equation.s_nodes)
        s0_index = np.abs(asset_grid - bsf_pricing_params['S0']).argmin()

        result1_at_s0 = result1[s0_index, 0]
        result2 = test_bsf.get_black_scholes_merton_price()

        print(result1_at_s0)
        print(result2)
        diff = abs(result1_at_s0 - result2)

        assert diff < 1e-6